import base64
import json
import logging
from django.db import models
from django.conf import settings
from django.contrib.auth.models import BaseUserManager, AbstractBaseUser, PermissionsMixin
from django.core.urlresolvers import reverse
from django.core.validators import validate_email
from django.template import Context
from django.template.loader import get_template
from django.utils.translation import ugettext as _
from django.core.exceptions import ObjectDoesNotExist, ValidationError

from exceptions import InvalidParameter
from yapi.utils import generate_key
from yutils.email import EmailMessage

# Instantiate logger.
logger = logging.getLogger(__name__)


class UserManager(BaseUserManager):
    """
    Since the custom User model defines different fields than Django's default, this custom
    manager must be implemented.
    """
    
    def create_user(self, name, email, password=None):
        """
        Creates and saves a User.
        """
        user = self.model(name=name, email=email)
        user.set_password(password)
        user.save()
        return user

    def create_superuser(self, name, email, password):
        """
        Creates and saves a superuser.
        """
        user = self.create_user(name, email, password=password)
        user.is_active = True
        user.is_staff = True
        user.is_superuser = True
        user.save()
        return user


class User(AbstractBaseUser, PermissionsMixin):
    """
    Custom User model.
    """
    # Fields.
    email = models.EmailField(verbose_name=_("Email address"), max_length=100, unique=True, db_index=True)
    name = models.CharField(verbose_name=_("Name"), max_length=100)
    is_active = models.BooleanField(verbose_name=_("Active"), help_text=_("Designates whether this user should be treated as active. Unselect this instead of deleting accounts."), default=False)
    is_staff = models.BooleanField(verbose_name=_("Staff status"), help_text=_("Designates whether the user can log into this admin site."), default=False)
    created_at = models.DateTimeField(verbose_name=_("Created at"), auto_now_add=True)

    # Required for custom User models.
    objects = UserManager()

    # The name of the field on the User model that is used as the unique identifier.
    USERNAME_FIELD = 'email'
    
    # A list of the field names that must be provided when creating a user via the createsuperuser management command.
    REQUIRED_FIELDS = ['name']

    def get_full_name(self):
        """
        A longer formal identifier for the user.
        """
        return self.name

    def get_short_name(self):
        """
        A short, informal identifier for the user.
        """
        return self.name

    def __unicode__(self):
        """
        String representation of the model instance.
        """
        return self.email
    
    def pending_activation(self):
        """
        Returns whether the account is pending authentication or not.
        """
        # The account is considered to be pending authentication if there is an
        # 'unused' activation key.
        try:
            ActivationKey.objects.get(user=self, activated_at__isnull=True)
            return True
        except ObjectDoesNotExist:
            return False
        
    @staticmethod
    def new(name, email, password):
        """
        Creates a new account.
        """
        # Validate email address.
        try:
            validate_email(email)
        except ValidationError:
            raise InvalidParameter('email', _("Please provide a valid email address"))
        
        # Check if account with given email already exists.
        try:
            User.objects.get(email=email)
            # If this place is reached, then account with given email already exists. Abort.
            raise InvalidParameter('email', _("This email is already registered."))
        except ObjectDoesNotExist:
            pass
        
        # All check, create account.
        user = User.objects.create_user(name=name, email=email, password=password)
        
        # Create activation key.
        try:
            ActivationKey.new(user)
        
        # Unable to create activation key, abort account creation.
        except:
            logger.error('Unable to create activation key! Name: ' + name + ', Email: ' + email, exc_info=1)
            user.delete()
            raise
        
        # Return.
        return user
    
    def update(self, name=None, password=None):
        """
        Update user account.
        """
        if name and name != '':
            self.name = name
        if password and password != '':
            self.set_password(password)
        self.save()
    
    
class UserPhoto(models.Model):
    """
    User profile picture.
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL)
    file = models.ImageField(upload_to='accounts/photos/')
    
    def __unicode__(self):
        """
        String representation of the instance.
        """
        return str(self.user)


class ActivationKey(models.Model):
    """
    Keys used to activate new accounts.
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL)
    key = models.CharField(max_length=200, unique=True)
    created_at = models.DateTimeField(auto_now_add=True)
    activated_at = models.DateTimeField(blank=True, null=True)
    
    def __unicode__(self):
        """
        String representation of the model instance.
        """
        return str(self.user)
    
    @staticmethod
    def new(user):
        """
        Creates an activation key for given user.
        """
        activation_key = ActivationKey(user=user, key=generate_key(salt=user.email))
        activation_key.save()
        activation_key.send_activation_link()
        return True
    
    def get_activation_token(self):
        """
        Returns a Base64 encoded string containing the account's email and activation key.
        """
        token = {
            'email': self.user.email,
            'operation': 'activation',
            'key': self.key
        }
        return base64.b64encode(json.dumps(token))
    
    def send_activation_link(self):
        """
        Sends the activation URL to the customers's email.
        """
        try:
            # Email variables.
            d = Context({
                'name': self.user.name,
                'activation_link': settings.HOST_URL + reverse('accounts:confirm') + '?t=' + self.get_activation_token()
            })
            
            # Render plaintext email template.
            plaintext = get_template('accounts/email/create_confirmation.txt')
            text_content = plaintext.render(d)
            
            # Render HTML email template.
            html = get_template('accounts/email/create_confirmation.html')
            html_content = html.render(d)
            
            # Email options.
            subject = _("New Account")
            from_email = settings.ACCOUNT_CREATION_MAIL_SENDER
            to = [{ 'name': self.user.name, 'email': self.user.email }]
            
            # Build message and send.
            email = EmailMessage(sender=from_email,
                                recipients=to,
                                subject=subject,
                                text_content=text_content,
                                html_content=html_content,
                                tags=['YU2 Account Confirm'])
            result = email.send()
            
            # Check if email wasn't sent.
            if not result['sent']:
                logger.error('Account Confirmation Email Not Sent! Result: ' + str(result['result']))
                raise
        except:
            logger.error('Unable to send confirmation email', exc_info=1)
            return False
        
        # Return great success.
        return True


class AuthenticationLog(models.Model):
    """
    Logs successful and failed logins.
    """
    # Available account statuses.
    ACCOUNT_STATUS = (
        ('active', 'Active'),
        ('disabled', 'Disabled'),
        ('does_not_exist', 'Does Not Exist'),
        ('pending_activation', 'Pending Activation')
    )
    
    # Fields.
    date = models.DateTimeField(auto_now_add=True)
    email = models.CharField(max_length=100)
    valid_credentials = models.BooleanField(default=False)
    account_status = models.CharField(max_length=20, choices=ACCOUNT_STATUS)
    success = models.BooleanField(default=False)
    ip_address = models.CharField(max_length=50)
    metadata = models.TextField(blank=True) # Additional data on the subject.
    
    def __unicode__(self):
        """
        String representation of the model instance.
        """
        return self.email
    
    @staticmethod
    def new(email, valid_credentials, account_status, success, ip_address, metadata):
        """
        Logs an authentication event.
        """
        # Build authentication log.
        authlog = AuthenticationLog(email=email,
                                    valid_credentials=valid_credentials,
                                    account_status=account_status,
                                    success=success,
                                    ip_address=ip_address,
                                    metadata=metadata)
        
        # Shave and return.
        authlog.save()
        return authlog


from django.conf.urls import patterns, url

from handlers import AccountHandler, AccountPhotoHandler
from yapi.resource import Resource

urlpatterns = patterns('',
    url(r'^/?$', Resource(AccountHandler), name='index'),
    url(r'^/photo/?$', Resource(AccountPhotoHandler), name='photo')
)


from django import forms
from django.utils.translation import ugettext as _


class UserForm(forms.Form):
    """
    Form for updating User.
    """
    name = forms.CharField(max_length=100)
    password_old = forms.CharField(max_length=100, required=False)
    password_new = forms.CharField(max_length=100, required=False)
    password_confirm = forms.CharField(max_length=100, required=False)
    
    def __init__(self,*args,**kwargs):
        """
        Constructor.
        """
        self.user = kwargs.pop('user')
        super(UserForm,self).__init__(*args, **kwargs)
    
    def clean(self):
        """
        Override default method to add additional validation.
        """
        # Superclass stuffz.
        cleaned_data = super(UserForm, self).clean()
        
        # New password provided.
        if cleaned_data.get('password_new'):
            
            # 1) Old password must check.
            if not cleaned_data.get('password_old'):
                self._errors['password_old'] = self.error_class([_('This field is required.')])
                del cleaned_data['password_new']
            elif not self.user.check_password(cleaned_data.get('password_old')):
                    self._errors['password_old'] = self.error_class([_('Invalid password.')])
                    del cleaned_data['password_new']
                    
            # 2) Double check new password with confirmation.
            elif not cleaned_data.get('password_confirm'):
                self._errors['password_confirm'] = self.error_class([_('This field is required.')])
                del cleaned_data['password_new']
            elif cleaned_data.get('password_confirm') != cleaned_data.get('password_new'):
                self._errors['password_confirm'] = self.error_class([_('Confirmation password doesn\'t match.')])
                del cleaned_data['password_new']
        
        # Return.
        return cleaned_data
    

class UserPhotoForm(forms.Form):
    """
    Form for User account photo.
    """
    file = forms.ImageField()


from django.conf.urls import patterns, url

import views

urlpatterns = patterns('',
                       
    url(r'^/?$', views.index, name='index'),
    url(r'^/login/?$', views.login_account, name='login'),
    url(r'^/logout/?$', views.logout_account, name='logout'),
    url(r'^/create/?$', views.create_account, name='create'),
    url(r'^/confirm/?$', views.confirm_account, name='confirm')
)


import base64
import json
import logging
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate, login, logout, get_user_model
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.http.response import HttpResponseRedirect, Http404
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from exceptions import InvalidParameter
from models import AuthenticationLog

# Instantiate logger.
logger = logging.getLogger(__name__)


@login_required
def index(request):
    """
    Account index.
    """
    # If user has photo.
    if hasattr(request.user, 'userphoto'):
        photo_url = request.user.userphoto.file.url
    else:
        photo_url = None
    
    # Render page.
    return render_to_response('accounts/index.html',
                              { 'name': request.user.name,
                                'email': request.user.email,
                                'photo_url': photo_url,
                                'change_photo_url': settings.HOST_URL + reverse('api:v1:accounts:photo') },
                              context_instance=RequestContext(request))


def login_account(request):
    """
    Login account.
    """
    # If a user is logged in, redirect to account page.
    if request.user.is_authenticated():
        return HttpResponseRedirect(reverse('accounts:index'))
    
    ################################################################
    #                     A form was submitted.                    #
    ################################################################
    if request.method == 'POST':
        
        # Additional information on the request that should be logged.
        metadata = { 'user_agent': request.META['HTTP_USER_AGENT'] }
        
        #
        # 1) Fetch params.
        #
        email = request.POST.get('email', '')
        password = request.POST.get('password', '')
        remember = request.POST.get('remember', None)
        
        # Mandatory parameters not provided.
        if email == '' or password == '':
            # Message.
            messages.error(request, _("Please provide your email and password."))
        
        #
        # 2) Validate credentials.
        #
        else:
            
            user = authenticate(email=email, password=password)
            
            ######################
            # VALID credentials. #
            ######################
            if user is not None:
                
                #
                # a) Account is ACTIVE.
                #
                if user.is_active:
                    
                    # Login user.
                    login(request, user)
                    
                    # If NOT flaged to remember.
                    if not remember: 
                        request.session.set_expiry(0)
                        
                    # Log authentication.
                    AuthenticationLog.new(email=email,
                                          valid_credentials=True,
                                          account_status='active',
                                          success=True,
                                          ip_address=request.META['REMOTE_ADDR'],
                                          metadata=json.dumps(metadata))
                    
                    #
                    # *** If login successful, REDIRECT to profile page or to provided url. ***
                    #
                    return HttpResponseRedirect(request.POST.get('next', reverse('accounts:index')))
                
                #
                # b) Account is PENDING ACTIVATION.
                #
                elif user.pending_activation():
                    
                    # Log authentication.
                    AuthenticationLog.new(email=email,
                                          valid_credentials=True,
                                          account_status='pending_activation',
                                          success=False,
                                          ip_address=request.META['REMOTE_ADDR'],
                                          metadata=json.dumps(metadata))
                    
                    # Message.
                    messages.warning(request, mark_safe(_("You didn't activate your account. <a href=\"\">Resend confirmation email</a>.")))
                
                #
                # c) Account is DISABLED.
                #
                else:
                    
                    # Log authentication.
                    AuthenticationLog.new(email=email,
                                          valid_credentials=True,
                                          account_status='disabled',
                                          success=False,
                                          ip_address=request.META['REMOTE_ADDR'],
                                          metadata=json.dumps(metadata))
                    
                    # Message.
                    messages.warning(request, _("Your account is disabled."))
            
            ########################
            # INVALID credentials. #
            ########################
            else:
                
                # Check if account with given email address exists.
                try:
                    user = get_user_model().objects.get(email=email)
                    # Account exists, check status.
                    if user.is_active:
                        account_status = 'active'
                    elif user.pending_activation():
                        account_status = 'pending_activation'
                    else:
                        account_status = 'disabled'
                # There is no account with given email address.
                except ObjectDoesNotExist:
                    account_status = 'does_not_exist'
                
                # Log authentication.
                AuthenticationLog.new(email=email,
                                      valid_credentials=False,
                                      account_status=account_status,
                                      success=False,
                                      ip_address=request.META['REMOTE_ADDR'],
                                      metadata=json.dumps(metadata))
                
                # Message.
                messages.error(request, _("Invalid credentials."))
        
        #
        # 3) If this place is reached, then login was unsuccessful.
        # The login page is rendered with respective message and pre-filled fields.
        #
        return render_to_response('accounts/login.html',
                                  { 'email': email, 'next': request.POST.get('next', None) },
                                  context_instance=RequestContext(request))
    
    ################################################################
    #                      Render login page.                      #
    ################################################################
    else:
        return render_to_response('accounts/login.html',
                                  { 'next': request.GET.get('next', None) },
                                  context_instance=RequestContext(request))


def logout_account(request):
    """
    Logout.
    """
    if request.user.is_authenticated():
        logout(request)
    return HttpResponseRedirect(request.GET.get('next', reverse('accounts:index')))


def create_account(request):
    """
    Creates a new account.
    """
    # Check if this feature is enabled.
    if not settings.ACCOUNT_CREATION_AVAILABLE:
        raise Http404
    
    # If a user is logged in, redirect to account page.
    if request.user.is_authenticated():
        return HttpResponseRedirect(reverse('accounts:index'))
    
    ################################################################
    #                     A form was submitted.                    #
    ################################################################
    if request.method == 'POST':
        
        #
        # 1) Fetch params.
        #
        name = request.POST.get('name', '')
        email = request.POST.get('email', '')
        password = request.POST.get('password', '')
        
        # Mandatory parameters not provided.
        if name == '' or email == '' or password == '':
            messages.error(request, _("Please provide your name, email and a password."))
        
        #
        # 2) Create account.
        #
        else:
            
            # If all parameters check, then a new account is created.
            try:
                get_user_model().new(name=name, email=email, password=password)
                messages.success(request, _("An email was sent in order to confirm your account."))
                return HttpResponseRedirect(reverse('accounts:login'))
            
            # Invalid parameters.
            except InvalidParameter as e:
                messages.error(request, e.message)    
            
            # Unknown error.
            except:
                messages.error(request, _("Error creating account, please contact support."))
        
        #
        # 3) If this place is reached, then account creation form contained errors.
        # The sign-up page is rendered with respective message and pre-filled fields.
        #
        return render_to_response('accounts/create.html',
                                  { 'name': name, 'email': email, 'next': request.POST.get('next', None) },
                                  context_instance=RequestContext(request))
    
    ################################################################
    #                     Render sign-up page.                     #
    ################################################################
    else:
        return render_to_response('accounts/create.html',
                                  { 'next': request.GET.get('next', None) },
                                  context_instance=RequestContext(request))


def confirm_account(request):
    """
    Confirm email address for account activation or email address update.
    """
    #
    # Fetch token.
    #
    try:
        token = request.GET['t']
        if token == '':
            raise KeyError
    except KeyError:
        raise Http404
    
    #
    # Process token.
    #
    try:
        
        # A valid token is a base64 encoded string.
        confirm_data = json.loads(base64.b64decode(token))
        
        # Containing the account's email, confirmation scenario and respective key.
        try:
            email = confirm_data['email']
            operation = confirm_data['operation']
            key = confirm_data['key']
        except KeyError:
            logger.info('Invalid account confirmation DATA! Data: ' + json.dumps(confirm_data) + ', IP Address: ' + request.META['REMOTE_ADDR'])
            raise Http404
    
    # Unable to b64 decode.
    except TypeError:
        logger.info('Invalid BASE64 account confirmation token! Token: ' + token + ', IP Address: ' +  request.META['REMOTE_ADDR'])
        raise Http404
    
    # Error decoding JSON
    except ValueError:
        logger.info('Invalid JSON account confirmation token! Token: ' + token + ', IP Address: ' +  request.META['REMOTE_ADDR'])
        raise Http404
    
    #
    # Validate operation.
    #
    if operation == 'activation':
        
        # Verify activation key, by attempting to authenticate user using it.
        user = authenticate(email=email, activation_key=key)
        
        # If a user is returned, then the activation key checked out and the account activated.
        if user is not None:
            
            # Login user.
            login(request, user)
            
            # Redirect to account page.
            return HttpResponseRedirect(reverse('accounts:index'))
            
        # Invalid activation key.
        else:
            logger.info('Invalid activation key! Data: ' + json.dumps(confirm_data) + ', IP Address: ' + request.META['REMOTE_ADDR'])
            raise Http404
    
    # Invalid operation.
    else:
        logger.info('Invalid account confirmation OPERATION! Data: ' + json.dumps(confirm_data) + ', IP Address: ' + request.META['REMOTE_ADDR'])
        raise Http404


import Image
import logging
import os

from serializers import UserSerializer, UserPhotoSerializer
from yaccounts.forms import UserForm, UserPhotoForm
from yaccounts.models import UserPhoto
from yapi.authentication import SessionAuthentication, ApiKeyAuthentication
from yapi.response import HTTPStatus, Response

# Instantiate logger.
logger = logging.getLogger(__name__)


class AccountHandler:
    """
    API endpoint handler.
    """
    # HTTP methods allowed.
    allowed_methods = ['GET', 'PUT']
    
    # Authentication & Authorization.
    authentication = [SessionAuthentication, ApiKeyAuthentication]
    
    def get(self, request):
        """
        Process GET request.
        """
        return Response(request=request,
                        data=request.auth['user'],
                        serializer=UserSerializer,
                        status=HTTPStatus.SUCCESS_200_OK)
        
    def put(self, request):
        """
        Process PUT request.
        """
        # Populate form with provided data.
        form = UserForm(request.data, user=request.auth['user'])
        
        # Provided data is valid.
        if form.is_valid():
            
            # Update account.
            request.auth['user'].update(name=form.cleaned_data['name'],
                                        password=form.cleaned_data['password_new'])
            
            # Return.
            return Response(request=request,
                            data=request.auth['user'],
                            serializer=UserSerializer,
                            status=HTTPStatus.SUCCESS_200_OK)
            
        # Form didn't validate!
        else:
            return Response(request=request,
                            data={ 'message': 'Invalid parameters', 'parameters': form.errors },
                            serializer=None,
                            status=HTTPStatus.CLIENT_ERROR_400_BAD_REQUEST)
        
        
class AccountPhotoHandler:
    """
    API endpoint handler.
    """
    # HTTP methods allowed.
    allowed_methods = ['POST']
    
    # Authentication & Authorization.
    authentication = [SessionAuthentication, ApiKeyAuthentication]
    
    def post(self, request):
        """
        Process POST request.
        """
        # Populate form with provided data.
        form = UserPhotoForm(request.POST, request.FILES)
        
        # Provided data is valid.
        if form.is_valid():
            
            # If user has photo.
            if hasattr(request.auth['user'], 'userphoto'):
                user_photo = request.auth['user'].userphoto
                # 1) Delete old file.
                os.remove(user_photo.file.path)
                # 2) Update with new one.
                user_photo.file = form.cleaned_data['file']
                
            # If not, create it.
            else:
                user_photo = UserPhoto(user=request.auth['user'],
                                       file=form.cleaned_data['file'])
                
            # Save.
            user_photo.save()
            
            # Resize image (user photos are fixed to 140x140)
            try:
                im = Image.open(user_photo.file.path)
                im = im.resize((140, 140), Image.ANTIALIAS)
                im.save(user_photo.file.path, format='JPEG')
            except IOError:
                logger.error('Error resizing user photo! User: ' + str(request.auth['user'].email))
                raise
            
            # Return.
            return Response(request=request,
                            data=user_photo,
                            serializer=UserPhotoSerializer,
                            status=HTTPStatus.SUCCESS_201_CREATED)
            
        # Form didn't validate!
        else:
            return Response(request=request,
                            data={ 'message': 'Invalid parameters', 'parameters': form.errors },
                            serializer=None,
                            status=HTTPStatus.CLIENT_ERROR_400_BAD_REQUEST)


import logging

from yapi.serializers import BaseSerializer

# Instantiate logger.
logger = logging.getLogger(__name__)


class UserSerializer(BaseSerializer):
    """
    Adds methods required for instance serialization.
    """
        
    def to_simple(self, obj, user=None):
        """
        Please refer to the interface documentation.
        """
        # Build response.
        simple = {
            'email': obj.email,
            'name': obj.name,
            'last_login': obj.last_login.strftime("%Y-%m-%d %H:%M:%S"),
            'photo': None
        }
        
        # If user has photo.
        if hasattr(obj, 'userphoto'):
            simple['photo'] = {
                'url': obj.userphoto.file.url
            }
        
        # Return.
        return simple
    
    
class UserPhotoSerializer(BaseSerializer):
    """
    Adds methods required for instance serialization.
    """
    
    def to_simple(self, obj, user=None):
        """
        Please refer to the interface documentation.
        """
        simple = {
            'url': obj.file.url
        }        
        return simple